from abc import ABC
from typing import List, Tuple, Optional

import numpy as np

from sziszapangma.core.alignment.step_type import StepType
from sziszapangma.core.alignment.alignment_step import AlignmentStep
from sziszapangma.core.alignment.distance_matrix_calculator import \
    DistanceCalculator
from sziszapangma.core.alignment.step_words import StepWords
from sziszapangma.core.alignment.alignment_processing_step import \
    AlignmentProcessingStep
from sziszapangma.core.alignment.word import Word


class AlignmentCalculator(ABC):
    _distance_matrix_calculator: DistanceCalculator

    def __init__(self, distance_matrix_calculator: DistanceCalculator):
        self._distance_matrix_calculator = distance_matrix_calculator

    def convert_processing_steps_to_result(
        self,
        processing_steps: List[AlignmentProcessingStep],
    ) -> List[AlignmentStep]:
        return [
            AlignmentStep(step.step_type, step.step_words, step.step_cost)
            for step in processing_steps
        ]

    def _get_reference_indexes_per_steps(
        self,
        steps: List[AlignmentProcessingStep]
    ) -> List[int]:
        counter = 0
        indexes = []
        for step in steps:
            indexes.append(counter)
            if step.step_type.contain_reference_word():
                counter = counter + 1
        return indexes

    def get_distance_matrix_between_words(
        self,
        reference: List[Word],
        hypothesis: List[Word]
    ) -> np.ndarray:
        return self._distance_matrix_calculator.calculate_distance_matrix(
            reference, hypothesis)

    @staticmethod
    def _get_initialized_levenshtein_matrix(
        reference: List[Word],
        hypothesis: List[Word]
    ) -> Tuple[np.ndarray, List[List[Optional[AlignmentProcessingStep]]]]:

        # TODO: consider about remove distance_arr replaced by steps_arr
        reference_len = len(reference)
        hypothesis_len = len(hypothesis)
        distance_arr = np.zeros((reference_len + 1) * (hypothesis_len + 1)) \
            .reshape((reference_len + 1, hypothesis_len + 1))
        steps_arr = [
            [None for _ in range(hypothesis_len + 1)]
            for _ in range(reference_len + 1)
        ]

        # levenshtein initial
        for ref_index in range(reference_len + 1):
            distance_arr[ref_index][0] = ref_index
            step_words = StepWords(
                reference[ref_index - 1] if ref_index > 0 else None,
                None
            )
            steps_arr[ref_index][0] = AlignmentProcessingStep\
                .levenshtein_deletion(ref_index - 1, step_words)
        for hyp_index in range(hypothesis_len + 1):
            distance_arr[0][hyp_index] = hyp_index
            step_words = StepWords(
                None,
                hypothesis[hyp_index - 1] if hyp_index > 0 else None
            )
            steps_arr[0][hyp_index] = AlignmentProcessingStep\
                .levenshtein_insertion(hyp_index - 1, step_words)

        return distance_arr, steps_arr

    @staticmethod
    def _get_levenshtein_processing_step_cross(
        prev_cross_distance: float,
        step_words: StepWords,
        current_distance: float
    ) -> AlignmentProcessingStep:
        return AlignmentProcessingStep.levenshtein_correct(
            prev_cross_distance, step_words, 0) \
            if current_distance == 0 \
            else AlignmentProcessingStep.levenshtein_substitution(
            prev_cross_distance, step_words, current_distance)

    def get_levenshtein_embedding_based(
        self,
        reference: List[Word],
        hypothesis: List[Word],
        distance_matrix: np.ndarray
    ) -> Tuple[np.ndarray, List[List[AlignmentProcessingStep]]]:

        reference_len = len(reference)
        hypothesis_len = len(hypothesis)
        distance_arr, steps_arr = self._get_initialized_levenshtein_matrix(
            reference, hypothesis)

        for ref_index in range(reference_len):
            for hyp_index in range(hypothesis_len):
                step_words = StepWords(reference[ref_index],
                                       hypothesis[hyp_index])
                current_distance = distance_matrix[ref_index][hyp_index]
                prev_cross_distance = distance_arr[ref_index][hyp_index]

                cross_go_step = self._get_levenshtein_processing_step_cross(
                    prev_cross_distance, step_words, current_distance)
                insertion_step = AlignmentProcessingStep.levenshtein_insertion(
                    distance_arr[ref_index + 1][hyp_index], step_words)
                deletion_step = AlignmentProcessingStep.levenshtein_deletion(
                    distance_arr[ref_index][hyp_index + 1], step_words)

                best_step = min([cross_go_step, insertion_step, deletion_step],
                                key=lambda it: it.total_distance())

                distance_arr[ref_index + 1][hyp_index + 1] = \
                    best_step.total_distance()
                steps_arr[ref_index + 1][hyp_index + 1] = best_step

        return distance_arr, steps_arr

    def extract_steps_path(
        self,
        steps_matrix: List[List[AlignmentProcessingStep]]
    ) -> List[AlignmentProcessingStep]:
        x = len(steps_matrix) - 1
        y = len(steps_matrix[0]) - 1
        to_return = []
        while not (x == 0 and y == 0):
            current_step = steps_matrix[x][y]
            to_return.append(current_step)
            if current_step.step_type == StepType.DELETION:
                x = x - 1
            elif current_step.step_type == StepType.INSERTION:
                y = y - 1
            else:  # creation and substitution
                y = y - 1
                x = x - 1
        return to_return[::-1]

    def _calculate_steps_path(
        self,
        reference: List[Word],
        hypothesis: List[Word]
    ) -> List[AlignmentProcessingStep]:
        distance_between_words = self.get_distance_matrix_between_words(
            reference, hypothesis)
        _, steps_matrix = self.get_levenshtein_embedding_based(
            reference, hypothesis, distance_between_words)
        return self.extract_steps_path(steps_matrix)

    def calculate_alignment(
        self,
        reference: List[Word],
        hypothesis: List[Word]
    ) -> List[AlignmentStep]:
        steps_path = self._calculate_steps_path(reference, hypothesis)
        return self.convert_processing_steps_to_result(steps_path)

    def calculate_alignment_weighted(
        self,
        reference: List[Word],
        hypothesis: List[Word],
        weights: List[float]
    ) -> List[AlignmentStep]:
        steps_path = self._calculate_steps_path(reference, hypothesis)
        return self.convert_processing_steps_to_result(steps_path)
