from abc import ABC, abstractmethod
from typing import List

import numpy as np

from sziszapangma.core.transformer.embedding_transformer import \
    EmbeddingTransformer
from sziszapangma.core.alignment.word import Word


class DistanceCalculator(ABC):
    @abstractmethod
    def calculate_distance_matrix(
        self,
        reference: List[Word],
        hypothesis: List[Word]
    ) -> np.array:
        pass

    @abstractmethod
    def calculate_distance_for_words(self, word1: Word, word2: Word) -> float:
        pass


class BinaryDistanceCalculator(DistanceCalculator):
    def calculate_distance_for_words(self, word1: Word, word2: Word) -> float:
        return 0 if word1.value == word2.value else 1

    def calculate_distance_matrix(
        self,
        reference: List[Word],
        hypothesis: List[Word]
    ) -> np.array:
        return np.array([
            [self.calculate_distance_for_words(reference_word, hypothesis_word)
             for hypothesis_word in hypothesis]
            for reference_word in reference
        ])


class CosineDistanceCalculator(DistanceCalculator):
    _embedding_transformer: EmbeddingTransformer

    def __init__(self, embedding_transformer: EmbeddingTransformer):
        self._embedding_transformer = embedding_transformer

    def calculate_distance_for_words(self, word1: Word, word2: Word) -> float:
        return self.cosine_distance_between_words_embeddings(
            self._embedding_transformer.get_embedding(word1.value),
            self._embedding_transformer.get_embedding(word2.value)
        )

    @staticmethod
    def cosine_distance_between_words_embeddings(
        word1_embedding: np.array,
        word2_embedding: np.array
    ) -> float:
        a = word1_embedding
        b = word2_embedding
        if a.shape != b.shape:
            raise RuntimeError(
                "array {} shape not match {}".format(a.shape, b.shape))
        if a.ndim == 1:
            a_norm = np.linalg.norm(a)
            b_norm = np.linalg.norm(b)
        elif a.ndim == 2:
            a_norm = np.linalg.norm(a, axis=1, keepdims=True)
            b_norm = np.linalg.norm(b, axis=1, keepdims=True)
        else:
            raise RuntimeError("array dimensions {} not right".format(a.ndim))
        similarity = np.dot(a, b.T) / (a_norm * b_norm)
        dist = 1. - similarity
        return dist

    def calculate_distance_matrix(
        self,
        reference: List[Word],
        hypothesis: List[Word]
    ) -> np.array:
        embeddings_dict = self._embedding_transformer.get_embeddings(
            list(set(it.value for it in (reference + hypothesis)))
        )
        return np.array([[
            self.cosine_distance_between_words_embeddings(
                embeddings_dict[reference_word.value],
                embeddings_dict[hypothesis_word.value],
            )
            for hypothesis_word in hypothesis]
            for reference_word in reference
        ])
